/*
    tcl-fuse - Tcl FUSE Interface
    A Tcl interface to the Linux kernel's FUSE subsystem.

    Copyright (C) 2008  Alexandros Stergiakis

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

TODO:
    * New "fuse" subcommands: unmount, list, info
    * Multithreaded operation (-s switch)
    * Return output of FsCmd instead of printing it to stderr.
    * Do our own FsCmd argument parsing (Tcl-programmer friendly).
All these points will increase complexity considerably, without 
significant benefit.
*/

/*
 * tclfuse.c
 *
 */

#define FUSE_USE_VERSION 26

#include <sys/types.h> // struct stat
#include <sys/stat.h> // struct stat
#include <unistd.h> // struct stat
#include <strings.h> // bzero
#include <stdlib.h> // free
#include <errno.h>

#include <tcl.h>
#include <fuse.h>
#include <fuse/fuse_common.h>
#include <fuse/fuse_lowlevel.h>

#include "proxy.h"
#include "tclfuse.h"

/*
 * Function Prototypes
 */

static int FuseCmd(ClientData clientData, 
        Tcl_Interp *interp, 
        int objc, 
        Tcl_Obj *CONST objv[]);

static int FsCmd(ClientData clientData, 
        Tcl_Interp *interp, 
        int argc, 
        CONST char *argv[]);

static void FsCmd_CleanUp(ClientData clientData);

static struct fuse *tclfuse_setup(int argc, char *argv[],
        const struct fuse_operations *op,
        size_t op_size,
        char **mountpoint,
        int *multithreaded,
        int *foreground,
        int *fd,
        void *user_data);

static void tclfuse_teardown(struct fuse *fuse, char *mountpoint);

static void ChannelEventHandler(ClientData clientData, int mask);

/*
 * Function Bodies
 */

int Fuse_Init(Tcl_Interp *interp) {
    if (Tcl_InitStubs(interp, "8.5", 0) == NULL) {
        return TCL_ERROR;
    }

    Tcl_CreateObjCommand(interp, "fuse", FuseCmd, (ClientData) NULL, NULL);

    Tcl_PkgProvide(interp, "fuse", "1.0");
    return TCL_OK;
}

static int FuseCmd(ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
    static char* cmds[] = { "create", "version", NULL };

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "option ...");
        return TCL_ERROR;
    }

    int index;
    if (Tcl_GetIndexFromObj(interp, objv[1], cmds, "option", 0, &index) != TCL_OK)
        return TCL_ERROR;

    switch (index) {
        case 0: /* create */
        {
            static char* ops[] = { 
                "-getattr", "-readdir", "-open", "-read", "-readlink", "-rename", 
                "-mknod", "-mkdir", "-unlink", "-rmdir", "-symlink", "-link", 
                "-chmod", "-chown", "-truncate", "-write", "-statfs", "-flush", 
                "-release", "-fsync", "-setxattr", "-getxattr", "-listxattr", 
                "-removexattr", "-opendir", "-releasedir", "-fsyncdir", 
                "-destroy", "-access", "-create", "-ftruncate", "-fgetattr", 
                "-lock", "-utimens", "-bmap", "-userdata", NULL 
            };

            if (objc < 3) {
                Tcl_WrongNumArgs(interp, 2, objv, "name ?option? ...");
                return TCL_ERROR;
            }

            struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) ckalloc(sizeof(struct tclfuse_fsdata));
            bzero((void *) data, sizeof(struct tclfuse_fsdata));

            data->op = (struct fuse_operations *) ckalloc(sizeof(struct fuse_operations));
            bzero((void *) data->op, sizeof(struct fuse_operations));

            int index, i;
            for (i = 3; i < objc; i++) {
                if (Tcl_GetIndexFromObj(interp, objv[i], ops, "operator", i - 1, &index) != TCL_OK) {
                    ckfree((char *) data->op);
                    ckfree((char *) data);
                    return TCL_ERROR; // error message is set
                }

                if (i == (objc - 1)) {
                    Tcl_SetResult(interp, "missing value to parameter", TCL_STATIC);
                    ckfree((char *) data->op);
                    ckfree((char *) data);
                    return TCL_ERROR;
                }

                i++;
                switch (index) {
                    case 0: 
                        data->op->getattr = _getattr;
                        data->getattr = objv[i]; break;
                    case 1: 
                        data->op->readdir = _readdir;
                        data->readdir = objv[i]; break;
                    case 2: 
                        data->op->open = _open;
                        data->open = objv[i]; break;
                    case 3: 
                        data->op->read = _read;
                        data->read = objv[i]; break;
                    case 4: 
                        data->op->readlink = _readlink;
                        data->readlink = objv[i]; break;
                    case 5: 
                        data->op->rename = _rename;
                        data->rename = objv[i]; break;
                    case 6: 
                        data->op->mknod = _mknod;
                        data->mknod = objv[i]; break;
                    case 7: 
                        data->op->mkdir = _mkdir;
                        data->mkdir = objv[i]; break;
                    case 8: 
                        data->op->unlink = _unlink;
                        data->unlink = objv[i]; break;
                    case 9: 
                        data->op->rmdir = _rmdir;
                        data->rmdir = objv[i]; break;
                    case 10: 
                        data->op->symlink = _symlink;
                        data->symlink = objv[i]; break;
                    case 11: 
                        data->op->link = _link;
                        data->link = objv[i]; break;
                    case 12: 
                        data->op->chmod = _chmod;
                        data->chmod = objv[i]; break;
                    case 13: 
                        data->op->chown = _chown;
                        data->chown = objv[i]; break;
                    case 14: 
                        data->op->truncate = _truncate;
                        data->truncate = objv[i]; break;
                    case 15: 
                        data->op->write = _write;
                        data->write = objv[i]; break;
                    case 16: 
                        data->op->statfs = _statfs;
                        data->statfs = objv[i]; break;
                    case 17: 
                        data->op->flush = _flush;
                        data->flush = objv[i]; break;
                    case 18: 
                        data->op->release = _release;
                        data->release = objv[i]; break;
                    case 19: 
                        data->op->fsync = _fsync;
                        data->fsync = objv[i]; break;
                    case 20: 
                        data->op->setxattr = _setxattr;
                        data->setxattr = objv[i]; break;
                    case 21: 
                        data->op->getxattr = _getxattr;
                        data->getxattr = objv[i]; break;
                    case 22: 
                        data->op->listxattr = _listxattr;
                        data->listxattr = objv[i]; break;
                    case 23: 
                        data->op->removexattr = _removexattr;
                        data->removexattr = objv[i]; break;
                    case 24: 
                        data->op->opendir = _opendir;
                        data->opendir = objv[i]; break;
                    case 25: 
                        data->op->releasedir = _releasedir;
                        data->releasedir = objv[i]; break;
                    case 26: 
                        data->op->fsyncdir = _fsyncdir;
                        data->fsyncdir = objv[i]; break;
                    case 27: 
                        data->op->destroy = _destroy;
                        data->destroy = objv[i]; break;
                    case 28: 
                        data->op->access = _access;
                        data->access = objv[i]; break;
                    case 29: 
                        data->op->create = _create;
                        data->create = objv[i]; break;
                    case 30: 
                        data->op->ftruncate = _ftruncate;
                        data->ftruncate = objv[i]; break;
                    case 31: 
                        data->op->fgetattr = _fgetattr;
                        data->fgetattr = objv[i]; break;
                    case 32: 
                        data->op->lock = _lock;
                        data->lock = objv[i]; break;
                    case 33: 
                        data->op->utimens = _utimens;
                        data->utimens = objv[i]; break;
                    case 34: 
                        data->op->bmap = _bmap;
                        data->bmap = objv[i]; break;
                    case 35: 
                        data->userdata = objv[i]; break;
                }
            }

            for (i = 4; i < objc; i += 2)
                Tcl_IncrRefCount(objv[i]);

            char *name;
            name = Tcl_GetString(objv[2]);

            data->interp = interp;
            data->fsname = objv[2];
            Tcl_IncrRefCount(objv[2]);

            Tcl_InitHashTable(&data->mounts, TCL_STRING_KEYS);

            data->cmdtoken = Tcl_CreateCommand(interp, name, FsCmd, (ClientData) data, FsCmd_CleanUp);

            break; /* create */
        }

        case 1: /* version */
        {
            if (objc != 2) {
                Tcl_WrongNumArgs(interp, 2, objv, NULL);
                return TCL_ERROR;
            }

            Tcl_Obj *result = Tcl_NewObj();
            Tcl_SetIntObj(result, fuse_version());
            Tcl_SetObjResult(interp, result);

            break; /* version */
        }

    } /* switch */

    return TCL_OK;
}

static int FsCmd(ClientData clientData, Tcl_Interp *interp, int argc, CONST char *argv[]) {
    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) clientData;

    struct fuse *fuse;
    char *mountpoint;
    int multithreaded; /* NOTE:
                          There are some complexities due to Tcl's
                          threading model: An interpreter cannot be shared
                          among different threads, therefore we have to
                          create exact replicates of the original interpreter
                          (where the handlers for the FUSE operations are
                          loaded), for every new FUSE operation. Creating such
                          replicas is not a Tcl feature at the moment.

                          There is also the efficiency parameter:
                          It's not efficient to initiate a new Tcl 
                          Thread for every file system I/O request.
                          FUSE is using POSIX threads which are more
                          efficient. Replicating the exact environment
                          of the original would be even more time
                          consuming.
                       */
    int foreground;
    int fd;

    /* NOTE:
       We use the argument parsing functions of FUSE to parse the Tcl command arguments. 
       Doing so, we reuse code and simplify the design, the only drawback is that FUSE lib
       is sending output to stderr, which is not returned by the Tcl command as usual. */
    fuse = tclfuse_setup(argc, argv, data->op, sizeof(struct fuse_operations), &mountpoint,
            &multithreaded, &foreground, &fd, (void *) data);
    if (fuse == NULL)  { // NOTE: --help also causes: fuse == NULL
        // Error output is sent to stderr already.
        return TCL_ERROR;
    }

    if(foreground) {
        int res;
        res = fuse_loop(fuse);

        tclfuse_teardown(fuse, mountpoint);
        if (res == -1) {
            Tcl_SetResult(interp, "Failure encountered while serving FUSE event loop", TCL_STATIC);
            return TCL_ERROR;
        }
    } else {
        struct tclfuse_mpdata *mp = (struct tclfuse_mpdata *) ckalloc(sizeof(struct tclfuse_mpdata));

        Tcl_Channel chan = Tcl_MakeFileChannel(fd, TCL_READABLE);
        mp->chan = chan;
        mp->mountpoint = mountpoint;
        mp->fuse = fuse;
        mp->fsdata = data;

        int unique;
        Tcl_HashEntry *entryPtr;
        entryPtr = Tcl_CreateHashEntry(&data->mounts, mountpoint, &unique);
        if(! unique) {
            ckfree((char *) mp);
            Tcl_DetachChannel(NULL, chan);
            tclfuse_teardown(fuse, mountpoint);
            Tcl_SetResult(interp, "Mount point is in use", TCL_STATIC);
            return TCL_ERROR;
        }

        Tcl_SetHashValue(entryPtr, (ClientData) mp);

        data->mountcnt++;
        Tcl_CreateChannelHandler(mp->chan, TCL_READABLE, ChannelEventHandler, (void *) mp);
    }

    return TCL_OK;
}

static void ChannelEventHandler(ClientData clientData, int mask) {
    struct tclfuse_mpdata *mp = (struct tclfuse_mpdata *) clientData;
    struct fuse_session *se = fuse_get_session(mp->fuse);

    int res = 0;
    struct fuse_chan *ch = fuse_session_next_chan(se, NULL);
    size_t bufsize = fuse_chan_bufsize(ch);
    char *buf = (char *) ckalloc(bufsize);
    if (!buf) {
        Tcl_SetResult(mp->fsdata->interp, "fuse: failed to allocate read buffer", TCL_STATIC);
        Tcl_BackgroundError(mp->fsdata->interp);
        // NOTE: We don't teardown in this case, we just don't process the request.
        return;
    }

    while (!fuse_session_exited(se)) {
        struct fuse_chan *tmpch = ch;
        res = fuse_chan_recv(&tmpch, buf, bufsize);
        if (res == -EINTR)
            continue;
        if (res <= 0)
            break;
        fuse_session_process(se, buf, res, tmpch);
        break;
    }

    ckfree(buf);
    fuse_session_reset(se);

    if (fuse_session_exited(se) || res < 0) {
        if (res < 0) {
            Tcl_SetResult(mp->fsdata->interp, "fuse: error while receiving FUSE request", TCL_STATIC);
            Tcl_BackgroundError(mp->fsdata->interp);
        }

        mp->fsdata->mountcnt--;
        if(mp->fsdata->mountcnt <= 0) {
            Tcl_DeleteCommandFromToken(mp->fsdata->interp, mp->fsdata->cmdtoken);
        } else {
            Tcl_HashEntry *entryPtr;
            entryPtr = Tcl_FindHashEntry(&mp->fsdata->mounts, mp->mountpoint);
            Tcl_DeleteHashEntry(entryPtr);

            tclfuse_teardown(mp->fuse, mp->mountpoint);
            if (Tcl_Close(mp->fsdata->interp, mp->chan) == TCL_ERROR) {
                // Error message is set.
                Tcl_BackgroundError(mp->fsdata->interp);
                return;
            }
            ckfree((char *) mp);
        }
        return;
    }
}

static struct fuse *tclfuse_setup(int argc, char *argv[],
                      const struct fuse_operations *op,
                      size_t op_size,
                      char **mountpoint,
                      int *multithreaded,
                      int *foreground,
                      int *fd,
                      void *user_data)
{
    struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
    struct fuse_chan *ch;
    struct fuse *fuse;
    int res;

    res = fuse_parse_cmdline(&args, mountpoint, multithreaded, foreground);
    if (res == -1)
        return NULL;

    ch = fuse_mount(*mountpoint, &args);
    if (!ch) {
        fuse_opt_free_args(&args);
        goto err_free;
    }

    fuse = fuse_new(ch, &args, op, op_size, user_data);
    fuse_opt_free_args(&args);
    if (fuse == NULL)
        goto err_unmount;

    *fd = fuse_chan_fd(ch);

    return fuse;

err_unmount:
    fuse_unmount(*mountpoint, ch);
    if (fuse)
        fuse_destroy(fuse);
err_free:
    free(*mountpoint); // It is allocated with malloc by FUSE internally.
    return NULL;
}

static void tclfuse_teardown(struct fuse *fuse, char *mountpoint) {
    struct fuse_session *se = fuse_get_session(fuse);
    struct fuse_chan *ch = fuse_session_next_chan(se, NULL);
    fuse_unmount(mountpoint, ch);
    fuse_destroy(fuse);
    free(mountpoint);
}

static void FsCmd_CleanUp(ClientData clientData) {
    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) clientData;

    Tcl_HashSearch search;
    Tcl_HashEntry *entryPtr;
    struct tclfuse_mpdata *mp;

    entryPtr = Tcl_FirstHashEntry(&data->mounts, &search);
    while (entryPtr != NULL) {
        mp = (struct tclfuse_mpdata *) Tcl_GetHashValue(entryPtr);

        Tcl_DeleteHashEntry(entryPtr);

        tclfuse_teardown(mp->fuse, mp->mountpoint);
        if (Tcl_Close(data->interp, mp->chan) == TCL_ERROR) {
            // Error message is set.
            Tcl_BackgroundError(data->interp);
            return;
        }
        ckfree((char *) mp);

        //data->mountcnt--; // Not necessary, since we delete all.
        entryPtr = Tcl_NextHashEntry(&search);
    }
    Tcl_DeleteHashTable(&data->mounts);

    Tcl_DecrRefCount(data->fsname);
    Tcl_DecrRefCount(data->getattr);
    Tcl_DecrRefCount(data->readdir);
    Tcl_DecrRefCount(data->readlink);
    Tcl_DecrRefCount(data->rename);
    Tcl_DecrRefCount(data->mknod);
    Tcl_DecrRefCount(data->mkdir);
    Tcl_DecrRefCount(data->unlink);
    Tcl_DecrRefCount(data->rmdir);
    Tcl_DecrRefCount(data->symlink);
    Tcl_DecrRefCount(data->link);
    Tcl_DecrRefCount(data->chmod);
    Tcl_DecrRefCount(data->chown);
    Tcl_DecrRefCount(data->truncate);
    Tcl_DecrRefCount(data->open);
    Tcl_DecrRefCount(data->read);
    Tcl_DecrRefCount(data->write);
    Tcl_DecrRefCount(data->statfs);
    Tcl_DecrRefCount(data->flush);
    Tcl_DecrRefCount(data->release);
    Tcl_DecrRefCount(data->fsync);
    Tcl_DecrRefCount(data->setxattr);
    Tcl_DecrRefCount(data->getxattr);
    Tcl_DecrRefCount(data->listxattr);
    Tcl_DecrRefCount(data->removexattr);
    Tcl_DecrRefCount(data->opendir);
    Tcl_DecrRefCount(data->releasedir);
    Tcl_DecrRefCount(data->fsyncdir);
    Tcl_DecrRefCount(data->destroy);
    Tcl_DecrRefCount(data->access);
    Tcl_DecrRefCount(data->create);
    Tcl_DecrRefCount(data->ftruncate);
    Tcl_DecrRefCount(data->fgetattr);
    Tcl_DecrRefCount(data->lock);
    Tcl_DecrRefCount(data->utimens);
    Tcl_DecrRefCount(data->bmap);
    Tcl_DecrRefCount(data->userdata);

    ckfree((char *) data->op);
    ckfree((char *) data);
}
